#include <opencv2/opencv.hpp>

int main()
{
    std::string cascadePath("haarcascade_frontalface_alt.xml");
    cv::CascadeClassifier cascade(cascadePath);

    // Initialize Kalman filter.
    // Internal state has 6 elements (x, y, width, vx, vy, vw)
    // Measurement has 3 elements (x, y, width)
    // Note: Height = 1.3 x width, so it is not part of the state or measurement.
    cv::KalmanFilter KF(6, 3, 0);

    // Transition matrix is of the form
    // [
    //   1, 0, 0, dt, 0,  0,
    //   0, 1, 0, 0,  dt, 0,
    //   0, 0, 1, 0,  0,  dt,
    //   0, 0, 0, 1,  0,  0,
    //   0, 0, 0, 0,  1,  0,
    //   0, 0, 0, 0,  0,  1 
    // ]
    // because
    // x = x + vx * dt
    // y = y + vy * dt
    // w = y + vw * dt
    // vx = vx
    // vy = vy
    // vw = vw
    cv::setIdentity(KF.transitionMatrix);
    // KF.transitionMatrix.at<float>(3, 3) = 0.9F;
    // KF.transitionMatrix.at<float>(4, 4) = 0.9F;
    // KF.transitionMatrix.at<float>(5, 5) = 0.9F;
    std::cout << KF.transitionMatrix << std::endl;

    // Measurement matrix is of the form
    // [
    //  1, 0, 0, 0, 0, 0,
    //  0, 1, 0, 0, 0, 0,
    //  0, 0, 1, 0, 0, 0
    // ]
    // because we are detecting only x, y and w.
    // These quantities are updated.
    cv::setIdentity(KF.measurementMatrix);
    std::cout << KF.measurementMatrix << std::endl;

    // Variable to store detected x, y and w
    cv::Mat measurement(3, 1, CV_32FC1, cv::Scalar(0.0F));
    // Variables to store detected object and tracked object
    cv::Mat objectTracked(4, 1, CV_32FC1, cv::Scalar(0.0F));
    cv::Mat objectDetected(4, 1, CV_32FC1, cv::Scalar(0.0F));

    // Variables to store results of the predict and update (a.k.a correct step)
    cv::Mat updatedMeasurement(3, 1, CV_32FC1, cv::Scalar(0.0F));
    cv::Mat predictedMeasurement(6, 1, CV_32FC1, cv::Scalar(0.0F));

    // Variable to indicate measurement was updated
    bool measurementWasUpdated = false;

    cv::VideoCapture camera(0, cv::CAP_DSHOW);
    int64 ticks = cv::getTickCount(), preTicks;

    for (;;) {
        cv::Mat image, imageGray;
        camera >> image;
        if (image.empty()) break;

        // Update dt for transition matrix.
        // dt = time elapsed.
        preTicks = ticks;
        ticks = cv::getTickCount();
        float dt(float((ticks - preTicks) / cv::getTickFrequency()));
        KF.transitionMatrix.at<float>(0, 3) = dt;
        KF.transitionMatrix.at<float>(1, 4) = dt;
        KF.transitionMatrix.at<float>(2, 5) = dt;

        // predict
        predictedMeasurement = KF.predict();

        cv::cvtColor(image, imageGray, cv::COLOR_BGR2GRAY);

        bool detected = false;
        std::vector<cv::Rect> faceRect;
        cascade.detectMultiScale(imageGray, faceRect, 1.1, 1, 0, cv::Size(1, 1));
        cv::Rect objectDetected;
        for (cv::Rect &faceR : faceRect) {
            // confirm face
            if (0.45 < std::min(faceR.width, faceR.height) / std::max(faceR.width, faceR.height) && faceR.area() >= 10000) {
                if (!detected || faceR.area() > objectDetected.area()) {
                    objectDetected = faceR;
                    detected = true;
                    std::cout << faceR.area() << std::endl;
                }
            }
        }

        cv::Rect objectTracked;
        if (detected) {
            // Copy x, y, w from the detected rectangle
            measurement.at<float>(0, 0) = static_cast<float>(objectDetected.x);
            measurement.at<float>(1, 0) = static_cast<float>(objectDetected.y);
            measurement.at<float>(2, 0) = static_cast<float>(objectDetected.width);
            // Perform Kalman update step
            updatedMeasurement = KF.correct(measurement);
            // Use updated measurement if measurement was updated
            objectTracked.x = static_cast<int>(updatedMeasurement.at<float>(0, 0));
            objectTracked.y = static_cast<int>(updatedMeasurement.at<float>(1, 0));
            objectTracked.width = static_cast<int>(updatedMeasurement.at<float>(2, 0));
            objectTracked.height = static_cast<int>(1.3*updatedMeasurement.at<float>(2, 0));
        }
        else {
            // If measurement was not updated, use predicted values.
            objectTracked.x = static_cast<int>(predictedMeasurement.at<float>(0, 0));
            objectTracked.y = static_cast<int>(predictedMeasurement.at<float>(1, 0));
            objectTracked.width = static_cast<int>(predictedMeasurement.at<float>(2, 0));
            objectTracked.height = static_cast<int>(1.3*predictedMeasurement.at<float>(2, 0));
        }

        cv::rectangle(image, objectTracked, detected ? cv::Scalar(0, 255, 0) : cv::Scalar(0, 0, 255), 2);

        cv::imshow("Tracked face", image);

        if (cv::waitKey(10) == 27) break;
    }

    camera.release();
    cv::destroyAllWindows();
    return 0;
}
