import cv2
import numpy as np

cascadePath = "haarcascade_frontalface_alt.xml"
cascade = cv2.CascadeClassifier(cascadePath)

# Initialize Kalman filter.
# Internal state has 6 elements (x, y, width, vx, vy, vw)
# Measurement has 3 elements (x, y, width)
# Note: Height = 1.3 x width, so it is not part of the state or measurement.
KF = cv2.KalmanFilter(6, 3, 0)
# Transition matrix is of the form
# [ 
#   1, 0, 0, dt, 0,  0,
#   0, 1, 0, 0,  dt, 0,
#   0, 0, 1, 0,  0,  dt,
#   0, 0, 0, 1,  0,  0,
#   0, 0, 0, 0,  1,  0,
#   0, 0, 0, 0,  0,  1 
# ]
# because
# x = x + vx * dt
# y = y + vy * dt
# w = y + vw * dt
# vx = vx
# vy = vy
# vw = vw
KF.transitionMatrix = cv2.setIdentity(KF.transitionMatrix)
#KF.transitionMatrix[3,3] = 0.9
#KF.transitionMatrix[4,4] = 0.9
#KF.transitionMatrix[5,5] = 0.9
print(KF.transitionMatrix)
# Measurement matrix is of the form
# [
#  1, 0, 0, 0, 0, 0,
#  0, 1, 0, 0, 0, 0,
#  0, 0, 1, 0, 0, 0
# ]
# because we are detecting only x, y and w.
# These quantities are updated.
KF.measurementMatrix = cv2.setIdentity(KF.measurementMatrix)
print(KF.measurementMatrix)

# Variable to store detected x, y and w
measurement = np.zeros((3, 1), dtype=np.float32)
# Variables to store detected object and tracked object
objectTracked = np.zeros((4, 1), dtype=np.float32)
objectDetected = np.zeros((4, 1), dtype=np.float32)

# Variables to store results of the predict and update (a.k.a correct step)
updatedMeasurement = np.zeros((3, 1), dtype=np.float32)
predictedMeasurement = np.zeros((6, 1), dtype=np.float32)

# Variable to indicate measurement was updated
measurementWasUpdated = False

camera = cv2.VideoCapture(0,cv2.CAP_DSHOW)
ticks = cv2.getTickCount()

while True:
    hasFrame, image = camera.read() 
    if not hasFrame: 
        break
        
    # Update dt for transition matrix.
    # dt = time elapsed.
    preTicks = ticks;
    ticks = cv2.getTickCount()
    dt = (ticks - preTicks) / cv2.getTickFrequency()
    KF.transitionMatrix[0, 3] = dt
    KF.transitionMatrix[1, 4] = dt
    KF.transitionMatrix[2, 5] = dt
    # predict
    predictedMeasurement = KF.predict()
    
    imageGray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    detected = False
    faceRect = cascade.detectMultiScale(imageGray, scaleFactor=1.1, minNeighbors=1, minSize=(1,1))
    if len(faceRect) > 0:
        # confirm face
        for faceR in faceRect:
            faceSize = faceR[2]*faceR[3]
            if 0.45 >= min(faceR[2], faceR[3])/max(faceR[2], faceR[3]) or faceSize < 10000:
                continue
            if not detected:
                objectDetected = faceR
                detected = True
                print(faceSize)
            elif faceSize > objectDetected[2]*objectDetected[3]:
                objectDetected = faceR
                print(faceSize)
    
    if detected:
        # Copy x, y, w from the detected rectangle
        measurement = objectDetected[0:3].astype(np.float32)
        # Perform Kalman update step
        updatedMeasurement = KF.correct(measurement)
        # Use updated measurement if measurement was updated
        objectTracked[0:3, 0] = updatedMeasurement[0:3, 0].astype(np.int32)
        objectTracked[3, 0] = (1.3*updatedMeasurement[2, 0]).astype(np.int32)
    else:
        # If measurement was not updated, use predicted values.
        objectTracked[0:3, 0] = predictedMeasurement[0:3, 0].astype(np.int32)
        objectTracked[3, 0] = (1.3*predictedMeasurement[2, 0]).astype(np.int32)

    x, y, w, h = objectTracked.ravel()
    cv2.rectangle(image, (int(x), int(y)), (int(x+w), int(y+h)), (0,255,0) if detected else (0,0,255), 2)
    
    cv2.imshow("Tracked face",image)

    if cv2.waitKey(10) == 27:
        break

camera.release()    
cv2.destroyAllWindows()
